/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/***************************************************************************
 *   Copyright (C) 2008 by Tomas "ZmaTkar" Kosan                           *
 *   zmatkar@k25.cz                        	                           *
 *								           *
 *   Regulator pajeciho pera SL-NI                                         *
 *                                             .                           *
 ***************************************************************************/

#include <avr/interrupt.h>
#include <avr/signal.h>
#include <avr/sfr_defs.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h> 
#include <math.h>
#include <avr/eeprom.h>

// jedeme na 1MHz
#define F_CPU 1E6
#include <util/delay.h>
#include  "lcd.h"

// makra pro ovladani LED diod
#define LED_RED		PORTB &=~ _BV(PB3)
#define TEST_RED	PORTB ^= _BV(PB3)

#define LED_BLUE	PORTB &=~ _BV(PB4)
#define TEST_BLUE	PORTB ^= _BV(PB4)

#define LED_GREEN	PORTB &=~ _BV(PB5)
#define TEST_GREEN	PORTB ^= _BV(PB5)

#define LED_OFF		PORTB |= _BV(PB3) | _BV(PB4) | _BV(PB5);

// ovladani spinaciho FETu
#define SWITCH_FET_OFF	PORTB &=~ _BV(PB1)
#define SWITCH_FET_ON	PORTB |= _BV(PB1)

// nastaveni zakladnich parametru
// minimalni nastavitelna teplota
#define T_MIN		150
// odchylka od nastavene hodnoty, kdy jeste indikujeme
// stav pripraveno
#define T_DELTA		5
// pocet pulzu regulace
#define POWER_RES	128

// napeti pro vypnuti pajky
#define U_IDLE		930
// 5 min. idle timer
#define MAX_IDLE_TIME	2307

// prepinani vstupup AD prevodniku
#define TEMP_AMB	0xC0
#define U_N		0xC1
#define IDLE_SENSE	0xC2
#define TEMP_SOLD	0xC4
#define	U_REG		0xC5

// vypisy na displej
#define MSG_READY	"Status:ready\n"
#define MSG_HEAT	"Status:heating\n"
#define MSG_COOL	"Status:cooling\n"
#define MSG_IDLE	"Idle mode\n"

// globalni promenne
// reulacni/nastavovaci napeti
volatile uint32_t U_reg;
// napeti z clanku - teplota
volatile uint32_t T_sold;
// teplota okoli
volatile int32_t T_amb;
// napajeci napeti
volatile uint32_t U_n;
// napeti z cisla odlozeni pajky
volatile uint32_t U_idle;
// pomer sepnuti MOSFET tr.
volatile uint8_t power_ratio = 0;
// citac necinnosti
volatile uint32_t idle_counter = 0;
// priznak pro preklesni displaye
volatile uint8_t repaint_flag = 1;

// inicializace prommenych
void init_IO(void)
{
	// PB0 - LED, PB1 - ON/Off heating, PB2 - pulses for -Ucc generation
	DDRB |= _BV(PB0) | _BV(PB1) | _BV(PB2) | _BV(PB3) | _BV(PB4) | _BV(PB5);
	PORTB &=~ _BV(PB2); 
	DDRD |= _BV(PD5);
	PORTD |= _BV(PD5);

	DDRD &=~ _BV(PD6) | _BV(PD7);
	PORTD &=~ _BV(PD6) | _BV(PD7);

	LED_OFF;

	// generator PWM - B kanal vyrabi -Un pro OZ
	TCCR1B |= _BV(CS10) | _BV(WGM13) | _BV(WGM12);
	TCCR1A |= _BV(WGM11) | _BV(COM1B1);
	
	ICR1  = 200;
	OCR1A = 100; 
	OCR1B = 100;

	// interrupt every 0.26 s
	TCCR2 |=  _BV(CS22) | _BV(CS21) | _BV(CS20);
	OCR2 = 255;
	TIMSK |= _BV(OCIE2);

	// init ADC
	ADMUX = U_REG;
	ADCSRA = _BV(ADIE) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0) | _BV(ADEN) | _BV(ADFR);

	// init AC
	ACSR &=~ _BV(ACD);
	ACSR |= _BV(ACIE);	
}

/*
*	Softwareova implementace proporcialne - sumacniho (integracniho) regulatoru
*/
uint32_t PS_regulator(uint32_t T_x, uint32_t T_p)
{
	// 0.26s - tak casto volame tuto proceduru
	#define dT  0.26
	// konstanta PS prevodniku
	#define K_P 2
	// integracni konstanta
	#define T_R 180
	
	int32_t T_d;
	//int16_t K_t;
	int32_t ratio = 0;
	static double  S = 0;
	
	// T_d je vlstne epsilon - odchylka
	T_d = (T_p - T_x);
	
	// zakladni rovnice regulatoru
	ratio = K_P*T_d	+ S;
	
	// sumujeme (integrujeme)
	if ((ratio >= 0) && (ratio <= POWER_RES))
		S = S + (((double)K_P*T_d)/T_R); 

	// omezoac hodnoty ratio			
	if (ratio < 0)
		ratio = 0;
	if (ratio > POWER_RES)
		ratio = POWER_RES;

	// vypnuti, kdyz jsme dlouho idle
	if (idle_counter > MAX_IDLE_TIME)
		ratio = 0;

	return ratio;	
}


ISR(TIMER2_COMP_vect)
{
	repaint_flag = 1;
	if (U_idle > U_IDLE)
	{
		if (idle_counter < MAX_IDLE_TIME+1)
			idle_counter++;
	}
	else
		idle_counter = 0;
}

ISR(ANA_COMP_vect)
{
	ACSR &=~ _BV(ACIE);
	static uint8_t counter = 0;
	
	if (++counter > POWER_RES)
		counter = 0;

	if (counter < power_ratio)
		SWITCH_FET_ON;
	else
		SWITCH_FET_OFF;

	ACSR |= _BV(ACIE);
}

// samplovani hodnot z AD vstupu
ISR(ADC_vect, ISR_NOBLOCK)
{
	ADCSRA &=~ _BV(ADSC);
	uint16_t temp;
	temp = ADCW;

	switch (ADMUX)
	{
		case TEMP_AMB:
			U_reg = (U_reg + temp) >> 1;
			ADMUX = U_N;
			break;
		case U_N:
			T_amb = (T_amb + temp) >> 1;
			ADMUX = IDLE_SENSE;
			break;
		case IDLE_SENSE:
			U_n = (U_n + temp) >> 1;
			ADMUX = TEMP_SOLD;
			break;
		case TEMP_SOLD:
			U_idle = (U_idle + temp) >> 1;
			ADMUX = U_REG;
			break;
		case U_REG:
			T_sold = (T_sold+temp)/2;
			ADMUX = TEMP_AMB;
			break;
		default:
			ADMUX = TEMP_AMB;

	ADCSRA |= _BV(ADSC);
	}	
}

// vypocet teploty z napeti K clanku
double count_K_type(uint32_t int_value)
{
	double value = 28*(double)int_value;
	double tmp = value;
	double T_xf;
	T_xf =  2.508355E-2*(tmp);
	tmp = tmp*value;
	T_xf = T_xf + 7.860106E-8*(tmp); 
	tmp = tmp*value;
	T_xf = T_xf - 2.503131E-10*(tmp);
	tmp = tmp*value;
	T_xf = T_xf + 8.31527E-14*(tmp);
	tmp = tmp*value;
	T_xf = T_xf - 1.228034E-17*(tmp);
	tmp = tmp*value;
	T_xf = T_xf + 9.804036E-22*(tmp);
	tmp = tmp*value;
	T_xf = T_xf - 4.413030E-26*(tmp);
	tmp = tmp*value;
	T_xf = T_xf + 1.057734E-30*(tmp);
	tmp = tmp*value;
	T_xf = T_xf - 1.052755E-35*(tmp);
	return T_xf;
}


void main(void)
{
	char buffer_1 [10];
	char buffer_2 [10];
	char buffer_3 [10];
	int32_t T_min, T_delta;
	uint32_t T_p, T_x;
	int8_t K_t = 0;
	int8_t K2_t = 0;
	uint32_t t_K;
	uint8_t timer = 0;
	double T_xf;

	init_IO();

	lcd_init(LCD_DISP_ON);
	lcd_puts_P("Solder station\nVersion 0.5");
	_delay_ms(2000);
	lcd_clrscr();
	lcd_puts_P("Ing. Tomas Kosan\nzmatkar.k25.cz");
	_delay_ms(2000);
	lcd_clrscr();
	lcd_puts_P("Waiting for:\n-U_n settle");
	_delay_ms(500);
	lcd_puts_P(" OK");
	_delay_ms(500);

	SREG |= _BV ( SREG_I );

	ADCSRA |= _BV(ADSC);

	for (;;)
	{
		T_min = (T_amb>>1)-263;

		if ((T_p+5 < (U_reg/4)+T_MIN) || (T_p-5 > (U_reg/4)+T_MIN))
			idle_counter = 0;

		T_p = (U_reg/4)+T_MIN;

		T_x = (uint32_t)count_K_type(T_sold)+T_min;

		if (repaint_flag)
		{
			T_delta = T_p - T_x;
			power_ratio = PS_regulator(T_x, T_p);

//#define DEBUG
#ifdef DEBUG
			itoa(power_ratio, buffer_1, 10);
			//utoa(T_min+(uint32_t)count_K_type(T_sold), buffer_1, 10);
#endif
			ultoa(T_p, buffer_2, 10);
			ultoa(T_x, buffer_3, 10);

			lcd_clrscr();
#ifdef DEBUG
			lcd_puts_P("Debug:");
			lcd_puts(buffer_1);
#else
			lcd_puts_P("T=");
			lcd_puts(itoa(T_min, buffer_1, 10));
			lcd_putc(223);
			lcd_puts_P("C ");
			if (idle_counter > MAX_IDLE_TIME)
				lcd_puts_P("IDLE/OFF");
			else
				if (idle_counter == 0)
				{
					if ((T_delta <= T_DELTA) && (T_delta >= -T_DELTA)) lcd_puts_P("READY");
				}
				else
				{
					lcd_puts_P("IDLE:");
					lcd_puts(utoa((idle_counter*0.26/60), buffer_1, 10));
					lcd_puts_P("min");
				}	

#endif
			 

			lcd_puts_P("\nTp/x=");
			lcd_puts(buffer_2);
			lcd_putc(223);
			lcd_puts_P("C/");
			lcd_puts(buffer_3);
			lcd_putc(223);
			lcd_puts("C");

			LED_OFF;
			if (idle_counter < MAX_IDLE_TIME)
			{
				if (T_delta < -T_DELTA) LED_GREEN;
				
				if (T_delta > T_DELTA) LED_RED;
				
				if ((T_delta <= T_DELTA) && (T_delta >= -T_DELTA)) LED_BLUE;
			}

			repaint_flag = 0;
		}
	
	}
}